import 'dart:convert';
import 'package:dotup_dart_geolocation_core/dotup_dart_geolocation_core.dart';
import 'package:dotup_dart_utils/dotup_dart_utils.dart';
import 'package:http/http.dart' as http;
import 'ApiResults/GeoCodeResponse.dart';
import 'HereMapsError.dart';
import 'MatchLevel.dart';
import 'PopularPlacesCategory.dart';
import 'ReverseGeoCodeMode.dart';

const Map<String, String> _categoryMap = {
  'eatAndDrink': 'eat-drink',
  'goingOut': 'going-out',
  'sightsMuseums': 'sights-museums',
  'atmBankExchange': 'atm-bank-exchange',
  'petrolStation': 'petrol-station'
};

class GetAddressOptions {
  const GetAddressOptions({
    this.mode = ReverseGeoCodeMode.retrieveAddresses,
    this.maxResults = 1,
    this.matchLevel,
  });

  final ReverseGeoCodeMode mode;
  final int maxResults;
  final MatchLevel? matchLevel;
}

class HereMapsGeoLocationService implements IGeoLocationService {
  final String apiKey;

  HereMapsGeoLocationService({required this.apiKey});

  /// discoverPlaces returns a List of places near the center of the [lat] and [lon] given
  /// [size] can be given to set the number of results returned by exploreNearbyPlaces
  /// By default the value of [size] is set to 20
  Future<Map<String, dynamic>> discoverPlaces({
    required double lat,
    required double lon,
    int size = 20,
  }) async {
    assert(size >= 0, "offset can't be negative");

    final headers = _createHeader();
    final body = _createMap()
      ..addAll(
        {
          'at': '$lat,$lon',
          'size': size.toString(),
          'tf': 'plain',
        },
      );

    var uri = Uri.https('places.ls.hereapi.com', '/places/v1/discover/here', body);
    return _callHereApi(uri, headers);
  }

  /// exploreNearbyPlaces returns a List of popular places near the center of the [lat] and [lon] given
  /// [category] can be given to retrieve the popular places of particular [category]
  /// [offset] can be given to set the number of results returned by exploreNearbyPlaces
  /// By default the value of [offset] is set to 20
  /// [nextUrl] can be given to fetch the next set of results

  Future<Map<String, dynamic>> explorePopularPlaces({
    required double lat,
    required double lon,
    PopularPlacesCategory? category,
    int offset = 20,
  }) async {
    assert(offset >= 0, "offset can't be negative");

    final headers = _createHeader();
    final data = _createMap();

    data['at'] = '$lat,$lon';

    if (category != null) {
      final cat = EnumUtils.getName(category);
      data['cat'] = _categoryMap.containsKey(cat) ? _categoryMap[cat]! : cat;
    }

    data['size'] = offset.toString();
    data['tf'] = 'plain';

    var uri = Uri.https('places.ls.hereapi.com', '/places/v1/discover/explore', data);
    return _callHereApi(uri, headers);
  }

  /// geoCodingAutoComplete provides better results for address searches with fewer keystrokes
  /// [query] is the mandatory parameter. The Results is returned on the basis of [query]
  /// [maxResults] can be given to restrict the suggestions based on [query]
  /// The valid value of [maxResults] is between 1 to 10

  Future<Map<String, dynamic>> geoCodingAutoComplete({
    required String query,
    int maxResults = 1,
  }) async {
    assert(maxResults >= 1 && maxResults <= 10, 'maxResults must be between 1 and 10');

    final headers = _createHeader();
    final data = _createMap();

    data['query'] = query;
    data['maxResults'] = '$maxResults';

    var uri = Uri.https('autocomplete.geocoder.ls.hereapi.com', '/6.2/suggest.json', data);
    return _callHereApi(uri, headers);
  }

  /// geoCode can be used to retrieves the latitude, longitude and complete address details
  /// based on the [searchText] provided

  Future<GeoCodeResponse> geoCode({
    String? country,
    String? state,
    String? county,
    String? city,
    String? district,
    String? street,
    String? housenumber,
    String? postalCode,
    String? searchtext,
  }) async {
    assert(
        country != null ||
            state != null ||
            county != null ||
            city != null ||
            district != null ||
            street != null ||
            housenumber != null ||
            postalCode != null ||
            searchtext != null,
        'At least one argument is required.');

    final headers = _createHeader();
    final data = _createMap();

    if (country != null) data['country'] = country;
    if (state != null) data['state'] = state;
    if (county != null) data['county'] = county;
    if (city != null) data['city'] = city;
    if (district != null) data['district'] = district;
    if (street != null) data['street'] = street;
    if (housenumber != null) data['housenumber'] = housenumber;
    if (postalCode != null) data['postalCode'] = postalCode;
    if (searchtext != null) data['searchtext'] = searchtext;

    var uri = Uri.https('geocoder.ls.hereapi.com', '/6.2/geocode.json', data);
    final result = await _callHereApi(uri, headers);
    return GeoCodeResponse.fromJson(result);
  }

  @override
  Future<Address?> getAddress({
    required double longitude,
    required double latitude,
    dynamic options = const GetAddressOptions(),
  }) async {
    // :(
    assert(options is GetAddressOptions);

    final response = await reverseGeoCode(
      lat: latitude,
      lon: longitude,
      mode: options.mode,
      maxResults: options.maxResults,
    );

    if (response.geoCode?.view != null) {
      var result = response.geoCode!.view!.first.result!;
      result.sort((v1, v2) => _getMatchLevelIndex(v1.matchLevel) > _getMatchLevelIndex(v2.matchLevel) ? 0 : 1);
      print(result);

      if (options.matchLevel != null) {
        final levelName = EnumUtils.getName(options.matchLevel!);
        result.where((element) => element.matchLevel == levelName).toList();
      }

      return result.first.location.address;
    }

    return null;
  }

  int _getMatchLevelIndex(String matchLevel) {
    return EnumUtils.getEnumFromName<MatchLevel>(MatchLevel.values, matchLevel).index;
  }

  /// reverseGeoCode retrieves the first address around a specified [lat] & [long]
  /// using a 250 meter radius to retrieve the address.
  /// [maxResults] can be given to get the number of the Results
  /// By default value of [maxResults] is 1
  /// [mode] defines the Reverse GeCoding Mode
  /// [ReverseGeoCodeMode.retrieveAddresses] Search for the closest street address or addresses
  /// [ReverseGeoCodeMode.retrieveAreas]  Retrieve the administrative area information for the position provided in the request
  /// [ReverseGeoCodeMode.retrieveLandmarks] Search for landmarks like parks and lakes in the proximity provided in the request
  /// [ReverseGeoCodeMode.trackPosition] Retrieve street and address information based on a position and bearing
  /// [ReverseGeoCodeMode.retrieveAll] Search for streets, administrative areas and landmarks. This mode aggregates the results of the three different modes in one call
  /// https://developer.here.com/documentation/geocoder/dev_guide/topics/reading-geocoding-response.html
  Future<GeoCodeResponse> reverseGeoCode({
    double? lat,
    double? lon,
    GeoCoordinates? geoCoordinates,
    ReverseGeoCodeMode mode = ReverseGeoCodeMode.retrieveAddresses,
    int maxResults = 1,
  }) async {
    assert(geoCoordinates != null || (lat != null && lon != null), "You've to provide coordinates");
    assert(maxResults <= 1, 'maxResults should be greater than 1');

    final headers = _createHeader();
    final data = _createMap();

    data['prox'] = geoCoordinates == null ? '$lat,$lon' : '${geoCoordinates.latitude},${geoCoordinates.longitude}';
    data['mode'] = EnumUtils.getName(mode);

    var uri = Uri.https('reverse.geocoder.ls.hereapi.com', '/6.2/reversegeocode.json', data);
    final result = await _callHereApi(uri, headers);
    return GeoCodeResponse.fromJson(result);
  }

  Future<Map<String, dynamic>> _callHereApi(Uri uri, Map<String, String>? headers) async {
    print(uri);
    final response = await http.get(uri, headers: headers);

    if (response.statusCode != 200) {
      final message = json.decode(response.body)['Details'];
      throw HereMapsError(statusCode: response.statusCode, message: message);
    }

    return json.decode(response.body);
  }

  Map<String, String> _createMap() {
    return {'apiKey': apiKey};
  }

  Map<String, String> _createHeader() {
    return {'Accept': 'application/json'};
  }

  @override
  Future<AddressWithGeoLocation?> getAddressWithGeoLocation() {
    throw HereMapsError(statusCode: -1, message: 'Not supported!');
  }

  @override
  Future<GeoCoordinates?> getGeoCoordinates({double accuracy = 5, int retries = 5}) {
    throw HereMapsError(statusCode: -1, message: 'Not supported!');
  }

  @override
  Future<GeoCoordinates?> getGeoCoordinatesFromAddress(Address address) async {
    final result = await geoCode(
      city: address.city,
      country: address.country,
      //  county: address.cou
      housenumber: address.houseNumber,
      postalCode: address.zip,
      state: address.state,
      street: address.street,
    );

    return result.geoCode?.getGeoCoordinates();
  }

  @override
  Future<Address?> getAddressFromGeoCoordinates(GeoCoordinates location) async {
    final result = await getAddress(longitude: location.longitude, latitude: location.latitude);

    return result;
  }
}
